@@ -1,24 +1,12 @@
 /*
- * Copyright (C) 2015  Vishesh Handa <vhanda@kde.org>
- *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
- *
- */
+    SPDX-FileCopyrightText: 2015 Vishesh Handa <vhanda@kde.org>
+
+    SPDX-License-Identifier: LGPL-2.1-or-later
+*/
 
 #include "fileindexscheduler.h"
 
+#include "baloodebug.h"
 #include "firstrunindexer.h"
 #include "newfileindexer.h"
 #include "modifiedfileindexer.h"
@@ -26,24 +14,30 @@
 #include "filecontentindexer.h"
 #include "filecontentindexerprovider.h"
 #include "unindexedfileindexer.h"
+#include "indexcleaner.h"
 
 #include "fileindexerconfig.h"
 
-#include <QTimer>
-#include <QDebug>
+#include <memory>
+
 #include <QDBusConnection>
 
 using namespace Baloo;
 
-FileIndexScheduler::FileIndexScheduler(Database* db, FileIndexerConfig* config, QObject* parent)
+FileIndexScheduler::FileIndexScheduler(Database* db, FileIndexerConfig* config, bool firstRun, QObject* parent)
     : QObject(parent)
     , m_db(db)
     , m_config(config)
     , m_provider(db)
     , m_contentIndexer(nullptr)
-    , m_indexerState(Idle)
-    , m_timeEstimator(config, this)
+    , m_indexerState(Startup)
+    , m_timeEstimator(this)
     , m_checkUnindexedFiles(false)
+    , m_checkStaleIndexEntries(false)
+    , m_isGoingIdle(false)
+    , m_isSuspended(false)
+    , m_isFirstRun(firstRun)
+    , m_inStartup(true)
 {
     Q_ASSERT(db);
     Q_ASSERT(config);
@@ -53,11 +47,15 @@ FileIndexScheduler::FileIndexScheduler(Database* db, FileIndexerConfig* config,
     connect(&m_powerMonitor, &PowerStateMonitor::powerManagementStatusChanged,
             this, &FileIndexScheduler::powerManagementStatusChanged);
 
-    m_contentIndexer = new FileContentIndexer(m_config, &m_provider, this);
+    if (m_powerMonitor.isOnBattery()) {
+        m_indexerState = LowPowerIdle;
+    }
+
+    m_contentIndexer = new FileContentIndexer(m_config->maxUncomittedFiles(), &m_provider, m_indexFinishedFiles, this);
     m_contentIndexer->setAutoDelete(false);
     connect(m_contentIndexer, &FileContentIndexer::done, this,
-            &FileIndexScheduler::scheduleIndexing);
-    connect(m_contentIndexer, &FileContentIndexer::newBatchTime, &m_timeEstimator,
+            &FileIndexScheduler::runnerFinished);
+    connect(m_contentIndexer, &FileContentIndexer::committedBatch, &m_timeEstimator,
             &TimeEstimator::handleNewBatchTime);
 
     QDBusConnection::sessionBus().registerObject(QStringLiteral("/scheduler"),
@@ -66,18 +64,38 @@ FileIndexScheduler::FileIndexScheduler(Database* db, FileIndexerConfig* config,
 
 FileIndexScheduler::~FileIndexScheduler()
 {
+    m_contentIndexer->quit();
     m_threadPool.waitForDone(0); // wait 0 msecs
 }
 
+void FileIndexScheduler::startupFinished() {
+    m_inStartup = false;
+    QTimer::singleShot(0, this, &FileIndexScheduler::scheduleIndexing);
+}
+
 void FileIndexScheduler::scheduleIndexing()
 {
-    if (m_threadPool.activeThreadCount() || m_indexerState == Suspended) {
+    if (!isIndexerIdle()) {
+        return;
+    }
+    m_isGoingIdle = false;
+
+    if (m_isSuspended) {
+        if (m_indexerState != Suspended) {
+            m_indexerState = Suspended;
+            Q_EMIT stateChanged(m_indexerState);
+        }
         return;
     }
 
-    if (m_config->isInitialRun()) {
+    if (m_isFirstRun) {
+        if (m_inStartup) {
+            return;
+        }
+
+        m_isFirstRun = false;
         auto runnable = new FirstRunIndexer(m_db, m_config, m_config->includeFolders());
-        connect(runnable, &FirstRunIndexer::done, this, &FileIndexScheduler::scheduleIndexing);
+        connect(runnable, &FirstRunIndexer::done, this, &FileIndexScheduler::runnerFinished);
 
         m_threadPool.start(runnable);
         m_indexerState = FirstRun;
@@ -87,7 +105,7 @@ void FileIndexScheduler::scheduleIndexing()
 
     if (!m_newFiles.isEmpty()) {
         auto runnable = new NewFileIndexer(m_db, m_config, m_newFiles);
-        connect(runnable, &NewFileIndexer::done, this, &FileIndexScheduler::scheduleIndexing);
+        connect(runnable, &NewFileIndexer::done, this, &FileIndexScheduler::runnerFinished);
 
         m_threadPool.start(runnable);
         m_newFiles.clear();
@@ -98,7 +116,7 @@ void FileIndexScheduler::scheduleIndexing()
 
     if (!m_modifiedFiles.isEmpty()) {
         auto runnable = new ModifiedFileIndexer(m_db, m_config, m_modifiedFiles);
-        connect(runnable, &ModifiedFileIndexer::done, this, &FileIndexScheduler::scheduleIndexing);
+        connect(runnable, &ModifiedFileIndexer::done, this, &FileIndexScheduler::runnerFinished);
 
         m_threadPool.start(runnable);
         m_modifiedFiles.clear();
@@ -109,7 +127,7 @@ void FileIndexScheduler::scheduleIndexing()
 
     if (!m_xattrFiles.isEmpty()) {
         auto runnable = new XAttrIndexer(m_db, m_config, m_xattrFiles);
-        connect(runnable, &XAttrIndexer::done, this, &FileIndexScheduler::scheduleIndexing);
+        connect(runnable, &XAttrIndexer::done, this, &FileIndexScheduler::runnerFinished);
 
         m_threadPool.start(runnable);
         m_xattrFiles.clear();
@@ -118,7 +136,39 @@ void FileIndexScheduler::scheduleIndexing()
         return;
     }
 
-    if (m_provider.size() && !m_powerMonitor.isOnBattery()) {
+    // No housekeeping, no content indexing
+    if (m_powerMonitor.isOnBattery()) {
+        if (m_indexerState != LowPowerIdle) {
+            m_indexerState = LowPowerIdle;
+            Q_EMIT stateChanged(m_indexerState);
+        }
+        return;
+    }
+
+    if (m_inStartup) {
+        if (m_indexerState != Startup) {
+            m_indexerState = Startup;
+            Q_EMIT stateChanged(m_indexerState);
+        }
+        return;
+    }
+
+    // This has to be above content indexing, because there can be files that
+    // should not be indexed in the DB (i.e. if config was changed)
+    if (m_checkStaleIndexEntries) {
+        auto runnable = new IndexCleaner(m_db, m_config);
+        connect(runnable, &IndexCleaner::done, this, &FileIndexScheduler::runnerFinished);
+
+        m_threadPool.start(runnable);
+        m_checkStaleIndexEntries = false;
+        m_indexerState = StaleIndexEntriesClean;
+        Q_EMIT stateChanged(m_indexerState);
+        return;
+    }
+
+    m_indexPendingFiles = m_provider.size();
+    m_indexFinishedFiles = 0;
+    if (m_indexPendingFiles) {
         m_threadPool.start(m_contentIndexer);
         m_indexerState = ContentIndexing;
         Q_EMIT stateChanged(m_indexerState);
@@ -127,7 +177,7 @@ void FileIndexScheduler::scheduleIndexing()
 
     if (m_checkUnindexedFiles) {
         auto runnable = new UnindexedFileIndexer(m_db, m_config);
-        connect(runnable, &UnindexedFileIndexer::done, this, &FileIndexScheduler::scheduleIndexing);
+        connect(runnable, &UnindexedFileIndexer::done, this, &FileIndexScheduler::runnerFinished);
 
         m_threadPool.start(runnable);
         m_checkUnindexedFiles = false;
@@ -135,24 +185,49 @@ void FileIndexScheduler::scheduleIndexing()
         Q_EMIT stateChanged(m_indexerState);
         return;
     }
-    m_indexerState = Idle;
-    Q_EMIT stateChanged(m_indexerState);
+
+    if (m_indexerState != Idle) {
+        m_indexerState = Idle;
+        Q_EMIT stateChanged(m_indexerState);
+    }
 }
 
 static void removeStartsWith(QStringList& list, const QString& dir)
 {
-    QMutableListIterator<QString> it(list);
-    while (it.hasNext()) {
-        const QString& file = it.next();
-        if (file.startsWith(dir)) {
-            it.remove();
-        }
+    const auto tail = std::remove_if(list.begin(), list.end(),
+        [&dir](const QString& file) {
+            return file.startsWith(dir);
+        });
+    list.erase(tail, list.end());
+}
+
+static void removeShouldNotIndex(QStringList& list, FileIndexerConfig* config)
+{
+    const auto tail = std::remove_if(list.begin(), list.end(),
+        [config](const QString& file) {
+            return !config->shouldBeIndexed(file);
+        });
+    list.erase(tail, list.end());
+}
+
+void FileIndexScheduler::updateConfig()
+{
+    // Interrupt content indexer, to avoid indexing files that should
+    // not be indexed (bug 373430)
+    if (m_indexerState == ContentIndexing) {
+        m_contentIndexer->quit();
     }
+    removeShouldNotIndex(m_newFiles, m_config);
+    removeShouldNotIndex(m_modifiedFiles, m_config);
+    removeShouldNotIndex(m_xattrFiles, m_config);
+    m_checkStaleIndexEntries = true;
+    m_checkUnindexedFiles = true;
+    scheduleIndexing();
 }
 
 void FileIndexScheduler::handleFileRemoved(const QString& file)
 {
-    if (!file.endsWith('/')) {
+    if (!file.endsWith(QLatin1Char('/'))) {
         m_newFiles.removeOne(file);
         m_modifiedFiles.removeOne(file);
         m_xattrFiles.removeOne(file);
@@ -166,30 +241,27 @@ void FileIndexScheduler::handleFileRemoved(const QString& file)
 
 void FileIndexScheduler::powerManagementStatusChanged(bool isOnBattery)
 {
-    qDebug() << "Power state changed";
+    qCDebug(BALOO) << "Power state changed - onBattery:" << isOnBattery;
     if (isOnBattery && m_indexerState == ContentIndexing) {
-        qDebug() << "On battery stopping content indexer";
+        qCDebug(BALOO) << "On battery, stopping content indexer";
         m_contentIndexer->quit();
-        //TODO: Maybe we can add a special state for suspended due to being on battery.
-        m_indexerState = Idle;
-        Q_EMIT stateChanged(m_indexerState);
-    } else if (!isOnBattery) {
+    } else {
         scheduleIndexing();
     }
 }
 
 void FileIndexScheduler::setSuspend(bool suspend)
 {
+    m_isSuspended = suspend;
     if (suspend) {
-        qDebug() << "Suspending";
+        qCDebug(BALOO) << "Suspending";
         if (m_indexerState == ContentIndexing) {
             m_contentIndexer->quit();
+        } else {
+            scheduleIndexing();
         }
-        m_indexerState = Suspended;
-        Q_EMIT stateChanged(m_indexerState);
     } else {
-        qDebug() << "Resuming";
-        m_indexerState = Idle;
+        qCDebug(BALOO) << "Resuming";
         // No need to emit here we'll be emitting in scheduling
         scheduleIndexing();
     }
@@ -200,7 +272,13 @@ uint FileIndexScheduler::getRemainingTime()
     if (m_indexerState != ContentIndexing) {
         return 0;
     }
-    return m_timeEstimator.calculateTimeLeft(m_provider.size());
+    uint remainingFiles = m_indexPendingFiles - m_indexFinishedFiles;
+    return m_timeEstimator.calculateTimeLeft(remainingFiles);
+}
+
+void FileIndexScheduler::scheduleCheckUnindexedFiles()
+{
+    m_checkUnindexedFiles = true;
 }
 
 void FileIndexScheduler::checkUnindexedFiles()
@@ -209,6 +287,18 @@ void FileIndexScheduler::checkUnindexedFiles()
     scheduleIndexing();
 }
 
+void FileIndexScheduler::scheduleCheckStaleIndexEntries()
+{
+    m_checkStaleIndexEntries = true;
+}
+
+void FileIndexScheduler::checkStaleIndexEntries()
+{
+    m_checkStaleIndexEntries = true;
+    scheduleIndexing();
+}
+
+
 uint FileIndexScheduler::getBatchSize()
 {
     return m_config->maxUncomittedFiles();
